/*****************************************************************************
 * grab-dvb.c: support for Digital Video Broadcast
 *****************************************************************************
 * $Id: grab-dvb.c,v 1.9 2005/08/11 23:25:01 pingus77 Exp $
 *****************************************************************************
 * Program under GNU General Public License (see ../COPYING)
 *****************************************************************************/


#include "config.h"
#ifdef HAVE_DVB

#include "config.h"
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>
#include <fcntl.h>
#include <ctype.h>

#include "grab-file.h"
#include "divx.h"

#include <linux/dvb/frontend.h>
#include <linux/dvb/dmx.h>

#include "avcodec.h"
#include "avformat.h"

#define width_essai  640 // pCodecCtx->width        // FIXME
#define height_essai 480 // pCodecCtx->height      // FIXME
#define CHANNEL_FILE "/.xdtv/channels-dvb.conf"


//#define delay 0.5

#include "avilib.h"

extern int filenoaudioout;
extern int use_file_audio;

extern avi_t *   videofile ;
extern double audiostamp;

///      to be removed ? ///

extern void **buffers;
extern double delay; // delay between two images
extern clock_t t_start;
extern struct tms start;
extern pthread_t audio_thread;
extern pthread_mutex_t file_mutex;

extern char* audiobuf;
extern int audiobuflen;
extern int audio_buf_max_len;
extern int use_file_audio;
extern int filenoaudioout;

extern void video_set_capture_size(int w, int h);
extern int nbufs;

static AVFormatContext *pFormatCtx;
static AVInputFormat* fmt;
static AVPacket packet;
int frameFinished;
extern AVInputFormat mpegts_demux;


/**** Video **/

static    int             videoStream;

static    AVCodecContext  *pCodecCtxVideo;
static    AVCodec         *pCodecVideo;
static    AVFrame         *pFrameRGB;
static    int             numBytesRGB;
static    uint8_t         *bufferRGB;
extern AVCodec mpeg2video_decoder;

/**** Audio **/

static    int             audioStream;
static    AVCodecContext  *pCodecCtxAudio;
static    AVCodec         *pCodecAudio;
int 	  out_size_audio;

extern AVCodec mp2_decoder;
/*****************************/

extern char* avifilename;

//static int fd,muxd;
//static int fd;
//static FILE* muxd;



static char FRONTEND_DEV [80];
static char DEMUX_DEV [80];
static char DVR_DEV [80];

static unsigned long BUF_SIZE = 64 * 1024;
static unsigned long long total_bytes;



#define ERROR(x...)                                                     \
        do {                                                            \
                fprintf(stderr, "ERROR: ");                             \
                fprintf(stderr, x);                                     \
                fprintf (stderr, "\n");                                 \
        } while (0)

#define PERROR(x...)                                                    \
        do {                                                            \
                fprintf(stderr, "ERROR: ");                             \
                fprintf(stderr, x);                                     \
                fprintf (stderr, " (%s)\n", strerror(errno));		\
        } while (0)


typedef struct {
	char *name;
	int value;
} Param;

static const Param inversion_list [] = {
	{ "INVERSION_OFF", INVERSION_OFF },
	{ "INVERSION_ON", INVERSION_ON },
	{ "INVERSION_AUTO", INVERSION_AUTO }
};

static const Param bw_list [] = {
	{ "BANDWIDTH_6_MHZ", BANDWIDTH_6_MHZ },
	{ "BANDWIDTH_7_MHZ", BANDWIDTH_7_MHZ },
	{ "BANDWIDTH_8_MHZ", BANDWIDTH_8_MHZ }
};

static const Param fec_list [] = {
	{ "FEC_1_2", FEC_1_2 },
	{ "FEC_2_3", FEC_2_3 },
	{ "FEC_3_4", FEC_3_4 },
	{ "FEC_4_5", FEC_4_5 },
	{ "FEC_5_6", FEC_5_6 },
	{ "FEC_6_7", FEC_6_7 },
	{ "FEC_7_8", FEC_7_8 },
	{ "FEC_8_9", FEC_8_9 },
	{ "FEC_AUTO", FEC_AUTO },
	{ "FEC_NONE", FEC_NONE }
};

static const Param guard_list [] = {
	{"GUARD_INTERVAL_1_16", GUARD_INTERVAL_1_16},
	{"GUARD_INTERVAL_1_32", GUARD_INTERVAL_1_32},
	{"GUARD_INTERVAL_1_4", GUARD_INTERVAL_1_4},
	{"GUARD_INTERVAL_1_8", GUARD_INTERVAL_1_8}
};

static const Param hierarchy_list [] = {
	{ "HIERARCHY_1", HIERARCHY_1 },
	{ "HIERARCHY_2", HIERARCHY_2 },
	{ "HIERARCHY_4", HIERARCHY_4 },
	{ "HIERARCHY_NONE", HIERARCHY_NONE }
};

static const Param constellation_list [] = {
	{ "QPSK", QPSK },
	{ "QAM_128", QAM_128 },
	{ "QAM_16", QAM_16 },
	{ "QAM_256", QAM_256 },
	{ "QAM_32", QAM_32 },
	{ "QAM_64", QAM_64 }
};

static const Param transmissionmode_list [] = {
	{ "TRANSMISSION_MODE_2K", TRANSMISSION_MODE_2K },
	{ "TRANSMISSION_MODE_8K", TRANSMISSION_MODE_8K },
};


#define LIST_SIZE(x) sizeof(x)/sizeof(Param)


/* prototypes */

int fmts_xaw_to_ffmpegcolorspace[MAX_VIDEO_FMT]={
  [VIDEO_RGB24_BE]   = PIX_FMT_RGB24,
  [VIDEO_YUV420] = PIX_FMT_YUV420P,
//  [VIDEO_YUYV]  = PIX_FMT_YUV422,
//  [VIDEO_YUV422P] = PIX_FMT_YUV422P,
0
};
  
static struct STRTAB norms[] = {
  {0,   "PAL"},
  {1,  "NTSC"},
  {2, "SECAM"},
  {-1, NULL}
};

static struct STRTAB inputs[] = {
  {0,   "Television"},
  {-1, NULL}
};



static
int parse_param (int fd, const Param * plist, int list_size, int *param)
{
	char c;
	int character = 0;
	int index = 0;

	while (1) {
		if (read(fd, &c, 1) < 1)
			return -1;	/*  EOF? */

		if ((c == ':' || c == '\n')
		    && plist->name[character] == '\0')
			break;

		while (toupper(c) != plist->name[character]) {
			index++;
			plist++;
			if (index >= list_size)	 /*  parse error, no valid */
				return -2;	 /*  parameter name found  */
		}

		character++;
	}

	*param = plist->value;

	return 0;
}



static
int parse_int(int fd, int *val)
{
	char number[11];	/* 2^32 needs 10 digits... */
	int character = 0;

	while (1) {
		if (read(fd, &number[character], 1) < 1)
			return -1;	/*  EOF? */

		if (number[character] == ':' || number[character] == '\n') {
			number[character] = '\0';
			break;
		}

		if (!isdigit(number[character]))
			return -2;	/*  parse error, not a digit... */

		character++;

		if (character > 10)	/*  overflow, number too big */
			return -3;	/*  to fit in 32 bit */
	};

	*val = strtol(number, NULL, 10);

	return 0;
}


static
int find_channel(int fd, const char *channel)
{
	int character = 0;

	while (1) {
		char c;

		if (read(fd, &c, 1) < 1)
			return -1;	/*  EOF! */

		if (c == ':' && channel[character] == '\0')
			break;

		if (toupper(c) == toupper(channel[character]))
			character++;
		else
			character = 0;
	};

	return 0;
}


static
int try_parse_int(int fd, int *val, const char *pname)
{
	int err;

	err = parse_int(fd, val);

	if (err)
		ERROR("error while parsing %s (%s)", pname,
		      err == -1 ? "end of file" :
		      err == -2 ? "not a number" : "number too big");

	return err;
}


static
int try_parse_param(int fd, const Param * plist, int list_size, int *param,
		    const char *pname)
{
	int err;

	err = parse_param(fd, plist, list_size, param);

	if (err)
		ERROR("error while parsing %s (%s)", pname,
		      err == -1 ? "end of file" : "syntax error");

	return err;
}


int parse(const char *fname, const char *channel,
	  struct dvb_frontend_parameters *frontend, int *vpid, int *apid, int *spid)
{
	int fd;
	int err;

	if ((fd = open(fname, O_RDONLY | O_NONBLOCK)) < 0) {
		PERROR ("could not open file '%s'", fname);
		perror ("");
		return -1;
	}

	if (find_channel(fd, channel) < 0) {
		ERROR("could not find channel '%s' in channel list", channel);
		return -2;
	}

	if ((err = try_parse_int(fd, (int*)&frontend->frequency, "frequency")))
		return -3;

	if ((err = try_parse_param(fd,
				   inversion_list, LIST_SIZE(inversion_list),
				   (int *) &frontend->inversion,
				   "inversion")))
		return -4;

	if ((err = try_parse_param(fd, bw_list, LIST_SIZE(bw_list),
				   (int *) &frontend->u.ofdm.bandwidth,
				   "bandwidth")))
		return -5;

	if ((err = try_parse_param(fd, fec_list, LIST_SIZE(fec_list),
				   (int *) &frontend->u.ofdm.code_rate_HP,
				   "code_rate_HP")))
		return -6;

	if ((err = try_parse_param(fd, fec_list, LIST_SIZE(fec_list),
				   (int *) &frontend->u.ofdm.code_rate_LP,
				   "code_rate_LP")))
		return -7;

	if ((err = try_parse_param(fd, constellation_list,
				   LIST_SIZE(constellation_list),
				   (int *) &frontend->u.ofdm.constellation,
				   "constellation")))
		return -8;

	if ((err = try_parse_param(fd, transmissionmode_list,
				   LIST_SIZE(transmissionmode_list),
				   (int *) &frontend->u.ofdm.
				   transmission_mode,
				   "transmission_mode")))
		return -9;

	if ((err = try_parse_param(fd, guard_list, LIST_SIZE(guard_list),
				   (int *) &frontend->u.ofdm.
				   guard_interval, "guard_interval")))
		return -10;

	if ((err = try_parse_param(fd, hierarchy_list,
				   LIST_SIZE(hierarchy_list),
				   (int *) &frontend->u.ofdm.
				   hierarchy_information,
				   "hierarchy_information")))
		return -11;

	if ((err = try_parse_int(fd, vpid, "Video PID")))
		return -12;

	if ((err = try_parse_int(fd, apid, "Audio PID")))
		return -13;

	if ((err = try_parse_int(fd, spid, "SUB PID")))
		return -14;

	close(fd);

	return 0;
}



static
int set_pesfilter (int fd, int pid, dmx_pes_type_t type, int dvr)
{
        struct dmx_pes_filter_params pesfilter;

        if (pid <= 0 || pid >= 0x1fff)
                return 0;

        pesfilter.pid = pid;
        pesfilter.input = DMX_IN_FRONTEND;
        pesfilter.output = dvr ? DMX_OUT_TS_TAP : DMX_OUT_DECODER;
        pesfilter.pes_type = type;
        pesfilter.flags = DMX_IMMEDIATE_START;

        if (ioctl(fd, DMX_SET_PES_FILTER, &pesfilter) < 0) {
                PERROR ("ioctl(DMX_SET_PES_FILTER) for %s PID failed",
                        type == DMX_PES_AUDIO ? "Audio" :
                        type == DMX_PES_VIDEO ? "Video" : "??");
                return -1;
        }

        return 0;
}



static
int setup_frontend (int fe_fd, struct dvb_frontend_parameters *frontend)
{
	struct dvb_frontend_info fe_info;

	if (ioctl(fe_fd, FE_GET_INFO, &fe_info) < 0) {
		PERROR("ioctl FE_GET_INFO failed");
		return -1;
	}

	if (fe_info.type != FE_OFDM) {
		ERROR ("frontend device is not a OFDM (DVB-T) device");
		return -1;
	}

	printf ("tuning to %i Hz\n", frontend->frequency);

	if (ioctl(fe_fd, FE_SET_FRONTEND, frontend) < 0) {
		PERROR("ioctl FE_SET_FRONTEND failed");
		return -1;
	}

	return 0;
}


static
int check_frontend (int fe_fd)
{
	fe_status_t status;
	uint16_t snr, signal;
	uint32_t ber, uncorrected_blocks;

	do {
		ioctl(fe_fd, FE_READ_STATUS, &status);
		ioctl(fe_fd, FE_READ_SIGNAL_STRENGTH, &signal);
		ioctl(fe_fd, FE_READ_SNR, &snr);
		ioctl(fe_fd, FE_READ_BER, &ber);
		ioctl(fe_fd, FE_READ_UNCORRECTED_BLOCKS, &uncorrected_blocks);

		printf ("status %02x | signal %04x | snr %04x | "
			"ber %08x | unc %08x | ",
			status, signal, snr, ber, uncorrected_blocks);

		if (status & FE_HAS_LOCK){
			printf("FE_HAS_LOCK\n");
			printf("ok, frontend runing ...\n");
			return 0;
		}

		usleep(1000000);
		printf("\n");
	} while (1);

	return 0;
}


//static char* list_channels[]={"M6","W9","NT1","Canal +","TF1","NRJ12","TPS Star","France 2","France 4","France 5","Arte","LCP","France 3",NULL};


static void process_data(int dvrfd, int tsfd)
{
	uint8_t buf[BUF_SIZE];
	int bytes, b2;

	bytes = read(dvrfd, buf, sizeof(buf));
	if (bytes < 0) {
		perror("read");
		if (errno == EOVERFLOW)
			return;
		fprintf(stderr, "exiting due to read() error\n");
		exit(1);
	}
	total_bytes += bytes;
	b2 = write(tsfd, buf, bytes);
	if (b2 == -1) {
		perror("write");
		exit(1);
	} else if (b2 < bytes)
		fprintf(stderr, "warning: read %d, but wrote only %d bytes\n", bytes, b2);
	else
		fprintf(stderr, "got %d bytes (%llu total)\n", bytes, total_bytes);
}


static int add_filter(unsigned int pid, const unsigned char* dmxdev)
{
	int fd;
	struct dmx_pes_filter_params f;

	fd = open((const char*)dmxdev, O_RDONLY);
	if (fd == -1) {
		perror("cannot open demux device");
		return 1;
	}

	memset(&f, 0, sizeof(f));
	f.pid = (uint16_t) pid;
	f.input = DMX_IN_FRONTEND;
	f.output = DMX_OUT_TS_TAP;
	f.pes_type = DMX_PES_OTHER;
	f.flags   = DMX_IMMEDIATE_START;

	if (ioctl(fd, DMX_SET_PES_FILTER, &f) == -1) {
		perror("DMX_SET_PES_FILTER");
		return 1;
	}
	fprintf(stderr, "set filter for PID 0x%04x on fd %d\n", pid, fd);

	//FIXME: don't leak the fd
	return 0;
}




int mpeg_read_data(AVPicture* buf, int* decoded){
   int total_decoded_video=0;
   int total_decoded_audio=0;
   int len;
   while(av_read_frame(pFormatCtx, &packet)>=0)
    {
	if (debug)
		fprintf(stderr,"** reading new av packet, stream : %d\n",packet.stream_index);
        // Is this a packet from the video stream?
        if(packet.stream_index==videoStream)
        {

            // Decode video frame
            len=avcodec_decode_video(pCodecCtxVideo, (AVFrame*) buf, &frameFinished, packet.data, packet.size);
	    if (debug)
		    fprintf(stderr,"\tsize of video packet decoded : %d\n",len);
            if (len < 0) {
        	fprintf(stderr, "Error while decoding video \n");
		return -3;
            }
	    total_decoded_video+=len;
            // Did we get a video frame?
	    // we hope that we have decode the frame :-)

            if(frameFinished){
	      if (debug)
		    fprintf(stderr,"\tvideo frame completed\n");
	            *decoded=total_decoded_video;
	   	    return 1;
	    }
        }
        // Is this a packet from the audio stream?
	else if (packet.stream_index==audioStream){

		while (packet.size>0){
        /* NOTE: the audio packet can contain several frames */
		 len=avcodec_decode_audio(pCodecCtxAudio, (int16_t *)audiobuf , &out_size_audio, packet.data, packet.size);
		 
		 if (len < 0) {
        	        fprintf(stderr, "Error while decoding audio \n");
			break;
            	 }
	         if (debug)
		    fprintf(stderr,"\tsize of audio packet decoded : %d\n",len);

	         packet.data += len;
                 packet.size -= len;
		 total_decoded_audio+=len;
		 // TODO manage properly if many audio frames in 1 packet !!
		}
		if (debug)
		    fprintf(stderr,"\taudio frame decoded, return : %d\n",out_size_audio);

	        *decoded=out_size_audio;
		return 2;
	}

        // Free the packet that was allocated by av_read_frame
        av_free_packet(&packet);
    }
    fprintf(stderr,"mmmhh return 0 ...\n");
    return 0;
}

static int mpegreadframe(AVPicture* buf) {
  int r;
  struct timeval tv;
  int len=0;
  
  do {
    r=mpeg_read_data(buf,&len);
    if(r==2) {	 // audio data
      gettimeofday (&tv, NULL);
      audiostamp=tv.tv_usec/1000000.0+tv.tv_sec;
      
      audiobuflen=len;

      if(debug>=2) fprintf(stderr,"audiobuflen=%d\n",audiobuflen);
      if(audiobuflen>0) {
	unlok_lock_mutex();      
      }
    }
    if(r==0) {
      fprintf(stderr,"END OF MPEG FILE REACHED\n");
      return 0;
    }
    if(r==1 && debug)
	fprintf(stderr,"VIDEO FILE DECODED, I return len = %d\n",len);
  } while(r==2 || r==0);
  if(r!=1) return r; else return len;
}


static int
grab_open(struct device_t *device)
{
  	int i=0;

	struct dvb_frontend_parameters frontend_param;
	char *confname = NULL;
	char *homedir = getenv ("HOME");
	char *channel = "France 4";
	int adapter = 0, frontend = 0, demux = 0, fddvr=0, dvr = 0;
	int vpid, apid, spid ;
	int frontend_fd, audio_fd, video_fd, dvr_fd;

	snprintf (FRONTEND_DEV, sizeof(FRONTEND_DEV),
		  "/dev/dvb/adapter%i/frontend%i", adapter, frontend);

	snprintf (DEMUX_DEV, sizeof(DEMUX_DEV),
		  "/dev/dvb/adapter%i/demux%i", adapter, demux);

	snprintf (DVR_DEV, sizeof(DVR_DEV),
		  "/dev/dvb/adapter%i/dvr%i", adapter, fddvr);


	printf ("using '%s' and '%s'\n", FRONTEND_DEV, DEMUX_DEV);


	
	if (!confname)
	{
		if (!homedir)
			ERROR ("$HOME not set");
		confname = malloc (strlen(homedir) + strlen(CHANNEL_FILE) + 1);
		memcpy (confname, homedir, strlen(homedir));
		memcpy (confname + strlen(homedir), CHANNEL_FILE,
	        	strlen(CHANNEL_FILE) + 1);
	}

	memset(&frontend_param, 0, sizeof(struct dvb_frontend_parameters));

	if (parse (confname, channel, &frontend_param, &vpid, &apid ,&spid))
		return -1;


	if (!avifilename){ // if avifilename, we're just testing ffmpeg decompression ...
		if ((frontend_fd = open(FRONTEND_DEV, O_RDWR)) < 0) {
			PERROR ("failed opening '%s'", FRONTEND_DEV);
			return -1;
		}

		if (setup_frontend (frontend_fd, &frontend_param) < 0)
			return -1;


		printf ("video pid %04d, audio pid %04d\n", vpid, apid);

		if ((video_fd = open(DEMUX_DEV, O_RDWR)) < 0) {
			PERROR("failed opening '%s'", DEMUX_DEV);
			return -1;
		}

		if (set_pesfilter (video_fd, vpid, DMX_PES_VIDEO, dvr) < 0)
			return -1;

		if ((audio_fd = open(DEMUX_DEV, O_RDWR)) < 0) {
			PERROR("failed opening '%s'", DEMUX_DEV);
			return -1;
		}

		if (set_pesfilter (audio_fd, apid, DMX_PES_AUDIO, dvr) < 0)
			return -1;


		check_frontend (frontend_fd);
		
		if ((dvr_fd = open(DVR_DEV, O_RDONLY | O_NONBLOCK)) < 0) {
			PERROR("failed opening '%s'", DVR_DEV);
			return -1;
		}
		if (dvr==0){	
			fprintf(stderr,"getting data from logical device !\n");
			if (add_filter(vpid, (const unsigned char*)DEMUX_DEV) != 0){
				fprintf(stderr,"add filter failed!\n");	
				return 1;
			}
			if (add_filter(apid, (const unsigned char*)DEMUX_DEV) != 0){
				fprintf(stderr,"add filter failed!\n");	
				return 1;
			}
		}
	}
  if (videofile==NULL)
	  videofile=malloc(sizeof(avi_t));
  videofile->a_bits=16;
  videofile->a_fmt=0x0050;
  videofile->a_chans=2;

    // Register all formats and codecs
    av_register_all();
    // Open video file
    if (!avifilename)
	    avifilename=DVR_DEV;
    fprintf(stderr,"Trying to open %s\n",avifilename);

    fmt = av_find_input_format("mpegts");
    av_register_input_format(fmt);
    fprintf(stderr,"av input format : %s (%s)\n",fmt->name, fmt->long_name);

  /* TODO : idealement, il faudrait se servir de AVFormatParameters et le passe à av_open_input_file
   * pour eviter de mouliner à l'ouverture du fichier ....
   *
    AVFormatParameters ap;
    ap.mpeg2ts_raw=1;
    ap.mpeg2ts_compute_pcr=1;

#if LIBAVCODEC_BUILD >= 4754
    ap.time_base.num=25;
    ap.time_base.den=1;
#else
    ap->frame_rate =25;
    ap->frame_rate_base=1;
#endif
    ap.sample_rate=48000;
    ap.channels=2;
    ap.width=544;
    ap.height=576;
    ap.pix_fmt=PIX_FMT_YUV420P;
   // const char *device; 
   // ap.standard;

    ap.video_codec_id=CODEC_ID_MPEG2TS;
    ap.audio_codec_id=CODEC_ID_MP2;
    */

    if(av_open_input_file(&pFormatCtx,avifilename, fmt,0,NULL)!=0){
	    fprintf(stderr,"cannot open file : %s\n",avifilename);
        return -1; // Couldn't open file
    }

    fprintf(stderr,"OK,  %s opened !\n",avifilename);
    // Retrieve stream information
    if(av_find_stream_info(pFormatCtx)<0){
	    fprintf(stderr,"cannot find stream\n");
        return -1; 
    }

    //Dump information about file onto standard error
    dump_format(pFormatCtx,0,avifilename,0);

    //Find the first video stream
    videoStream=-1;
    audioStream=-1;

    for(i=0; i<pFormatCtx->nb_streams; i++){
#if LIBAVFORMAT_BUILD >= 4629
        if(pFormatCtx->streams[i]->codec->codec_type==CODEC_TYPE_VIDEO)
#else
        if(pFormatCtx->streams[i]->codec.codec_type==CODEC_TYPE_VIDEO)
#endif
        {
            videoStream=i;
	    break;
        }
    }
	
    //Find the last audio stream
    for(i=0; i<pFormatCtx->nb_streams; i++){
#if LIBAVFORMAT_BUILD >= 4629
        if(pFormatCtx->streams[i]->codec->codec_type==CODEC_TYPE_AUDIO)
#else
        if(pFormatCtx->streams[i]->codec.codec_type==CODEC_TYPE_AUDIO)
#endif
        {
            audioStream=i;
        }
    }

   if(videoStream==-1){
	fprintf(stderr,"didn't find a video stream\n");
        return -1; 
    }
    if(audioStream==-1){
	fprintf(stderr,"didn't find an audio stream\n");
        return -1; 
    }
    fprintf (stderr,"video stream : %d , audio stream : %d\n",
		    videoStream, audioStream);
#if LIBAVFORMAT_BUILD >= 4629
    // Get a pointer to the codec context for the video stream
    pCodecCtxVideo = pFormatCtx->streams[videoStream]->codec;
    // Get a pointer to the codec context for the audio stream
    pCodecCtxAudio = pFormatCtx->streams[audioStream]->codec;
#else
    // Get a pointer to the codec context for the video stream
    pCodecCtxVideo = &pFormatCtx->streams[videoStream]->codec;
    // Get a pointer to the codec context for the audio stream
    pCodecCtxAudio = &pFormatCtx->streams[audioStream]->codec;
#endif
    // Find the decoder for the video stream
    if (NULL == (pCodecVideo=avcodec_find_decoder(pCodecCtxVideo->codec_id))){
	fprintf(stderr, "cannot found video codec\n");
        return -1;
    }

    // Find the decoder for the audio stream
    if (NULL == (pCodecAudio=avcodec_find_decoder(pCodecCtxAudio->codec_id))){
	fprintf(stderr,"cannot found audio codec\n");
        return -1; 
    }

    // Open video codec
    if(avcodec_open(pCodecCtxVideo, pCodecVideo)<0){
    	fprintf (stderr,"dvb: cannot open video codec\n");
        return -1; // Could not open codec
    }
 
    // Open audio codec
    if(avcodec_open(pCodecCtxAudio, pCodecAudio)<0){
    	fprintf (stderr,"dvb: cannot open audio codec\n");
        return -1; // Could not open codec
    }

   // Hack to correct wrong frame rates that seem to be generated by some 
    // codecs
#if LIBAVCODEC_BUILD >= 4754
    if(pCodecCtxVideo->time_base.num>1000 && pCodecCtxVideo->time_base.den==1)
        pCodecCtxVideo->time_base.den=1000;

    if(pCodecCtxAudio->time_base.num>1000 && pCodecCtxAudio->time_base.den==1)
        pCodecCtxAudio->time_base.den=1000;
#else
    if(pCodecCtxVideo->frame_rate>1000 && pCodecCtxVideo->frame_rate==1)
        pCodecCtxVideo->frame_rate_base=1000;

    if(pCodecCtxAudio->frame_rate>1000 && pCodecCtxAudio->frame_rate==1)
        pCodecCtxAudio->frame_rate_base=1000;
#endif
    videofile->a_rate=pCodecCtxAudio->sample_rate;
    videofile->width=  pCodecCtxVideo->width;
    videofile->height= pCodecCtxVideo->height;
#if LIBAVCODEC_BUILD >= 4754
    videofile->fps_den=pCodecCtxVideo->time_base.num;
    videofile->fps_num=pCodecCtxVideo->time_base.den;
#else
    videofile->fps_den=pCodecCtxVideo->frame_rate;
    videofile->fps_num=pCodecCtxVideo->frame_rate_base;
#endif

    audio_buf_max_len=AVCODEC_MAX_AUDIO_FRAME_SIZE;
    // Allocate audio buffer
    if (NULL==(audiobuf=(char*) malloc(audio_buf_max_len))){
 	perror("malloc");
    }
   
    // Allocate an AVFrame structure for RGB Image 
    // to be removed soon
    if (NULL==(pFrameRGB=avcodec_alloc_frame())){
    	fprintf (stderr,"dvb: cannot allocate frame RGB\n");
        return -1;
    }

   fprintf (stderr,"dvb: set capture size to %ldx%ld\n",
		   videofile->width, videofile->height);
   video_set_capture_size(videofile->width, videofile->height);

    // Determine required buffer size and allocate buffer
    numBytesRGB= avpicture_get_size(PIX_FMT_RGB24,videofile->width, videofile->height);

    bufferRGB=malloc(numBytesRGB*sizeof(uint8_t));

    // Assign appropriate parts of buffer to image planes in pFrameRGB
    avpicture_fill((AVPicture *)pFrameRGB, bufferRGB, PIX_FMT_RGB24,
       videofile->width, videofile->height);

  nbufs = nbufs_default;
  if(nbufs > 3) nbufs = 3;
  buffers = (void **)(malloc(nbufs*sizeof(void*)));
  for(i=0;i<nbufs;i++){
    buffers[i]=avcodec_alloc_frame();
  }
  delay = (double)videofile->fps_den / videofile->fps_num;

  t_start = times(&start);
  
  init_semaphores_or_mutex();

  file_audio_start();

  return 0;
}

static int
grab_close ()
{
  int i;
  file_audio_stop();
  destroy_semaphores_or_mutex();
  
  free(videofile);
  videofile=NULL;
  for(i=0;i<nbufs;i++) 
	  av_free(buffers[i]);
  av_free(pFrameRGB);
  free(buffers);
  free(bufferRGB);
  free(audiobuf);
  return 0;
}


static int
grab_overlay (int x, int y, int width, int height, int format,
              struct OVERLAY_CLIP *oc, int count)
{
  return 0;
}


static void * get_img (video_fmt f, int width, int height)
{
  struct timeval tv;
  int format ;
  AVPicture* pict;
  format = fmts_xaw_to_ffmpegcolorspace[f];
  if(!format) return 0;

  if(width!=videofile->width) {
      fprintf(stderr,
	      "dvb_width=%ld requested_width=%d !\n",videofile->width,width);
      return NULL;
    }
  if(height!=videofile->height) {
    fprintf(stderr,
	    "dvb_height=%ld requested_height=%d !\n",videofile->height,height);
    return NULL;
  }
  img++; 
  
  if(img==nbufs){
	  img=0;
  }
  mpegreadframe(buffers[img]);

  pict=(AVPicture*) buffers[img];
  if (debug)
 	fprintf(stderr,"linesize [0]:%d [1]:%d [2]:%d [3]:%d \n",
			pict->linesize[0],pict->linesize[1],pict->linesize[2],pict->linesize[3]);

  img_convert((AVPicture *)pFrameRGB, format, (AVPicture*) (buffers[img]),pCodecCtxVideo->pix_fmt ,
		 pCodecCtxVideo->width,pCodecCtxVideo->height);

  gettimeofday (&tv, NULL);
  videostampmax=tv.tv_usec/1000000.0+tv.tv_sec;
  if(videostampmax<videostampmin+delay)
    	usleep(1000000*(delay-videostampmax+videostampmin));
  videostampmin=videostampmax;

  if (debug)
	  fprintf(stderr,"return a new image, frame finished : %d\n",frameFinished);
  return /*((AVPicture*) buffers[img])->data[0];// */pFrameRGB->data[0];
}

	
static int fmt_available(video_fmt f) {
    return fmts_xaw_to_ffmpegcolorspace[f]!=0;
}

/* ---------------------------------------------------------------------- */

static int
grab_tune (unsigned long freq)
{

	
  return 0;
}

static int
grab_tuned ()
{
  return 1;
}

static int
grab_input (int input, int norm)
{
  return 0;
}

static int
grab_picture (int color, int bright, int hue, int contrast)
{
  return 0;
}

static int
grab_audio (int mute, int volume, int *mode)
{
  return 0;
}

static int is525(int norm) {
  return 0;
}

static int issecam(int norm) {
  return 0;
}

static void* get_buf(int i){
  return NULL;
}

//////////////// SPECIAL FUNCTIONS .... ///


struct GRABBER grab_dvb = {
  "dvb",
  norms, inputs,
  grab_open,
  grab_close,
  grab_overlay,
  fmt_available,
  get_img,
  grab_tune,
  grab_tuned,
  grab_input,
  grab_picture,
  grab_audio,
  is525,
  issecam,
  get_buf
};
#endif

